/*
 * Work under Copyright. Licensed under the EUPL.
 * See the project README.md and LICENSE.txt for more information.
 */

package net.dries007.tfc.util;

import java.util.Arrays;
import java.util.Collection;
import java.util.Random;
import java.util.Set;
import java.util.stream.Collectors;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import com.google.common.base.Joiner;
import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.item.EntityItem;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.item.ItemStack;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.IBlockAccess;
import net.minecraft.world.World;
import net.minecraftforge.registries.IForgeRegistryEntry;

import net.dries007.tfc.api.types.Rock;
import net.dries007.tfc.api.util.TFCConstants;
import net.dries007.tfc.objects.blocks.BlockPeat;
import net.dries007.tfc.objects.blocks.BlocksTFC;
import net.dries007.tfc.objects.blocks.stone.BlockRockVariant;

public final class Helpers
{
    private static final Joiner JOINER_DOT = Joiner.on('.');

    public static void spreadGrass(World world, BlockPos pos, IBlockState us, Random rand)
    {
        if (world.func_175671_l(pos.func_177984_a()) < 4 && world.func_180495_p(pos.func_177984_a()).getLightOpacity(world, pos.func_177984_a()) > 2)
        {
            if (us.func_177230_c() instanceof BlockPeat)
            {
                //noinspection ConstantConditions
                world.func_175656_a(pos, BlocksTFC.PEAT.func_176223_P());
            }
            else if (us.func_177230_c() instanceof BlockRockVariant)
            {
                BlockRockVariant block = ((BlockRockVariant) us.func_177230_c());
                world.func_175656_a(pos, block.getVariant(block.type.getNonGrassVersion()).func_176223_P());
            }
        }
        else
        {
            if (world.func_175671_l(pos.func_177984_a()) < 9) return;

            for (int i = 0; i < 4; ++i)
            {
                BlockPos target = pos.func_177982_a(rand.nextInt(3) - 1, rand.nextInt(5) - 3, rand.nextInt(3) - 1);
                if (world.func_189509_E(target) || !world.func_175667_e(target)) return;
                BlockPos up = target.func_177982_a(0, 1, 0);

                IBlockState current = world.func_180495_p(target);
                if (!BlocksTFC.isSoil(current) || BlocksTFC.isGrass(current)) continue;
                if (world.func_175671_l(up) < 4 || world.func_180495_p(up).getLightOpacity(world, up) > 3)
                    continue;

                if (current.func_177230_c() instanceof BlockPeat)
                {
                    //noinspection ConstantConditions
                    world.func_175656_a(target, BlocksTFC.PEAT_GRASS.func_176223_P());
                }
                else if (current.func_177230_c() instanceof BlockRockVariant)
                {
                    Rock.Type spreader = Rock.Type.GRASS;
                    if ((us.func_177230_c() instanceof BlockRockVariant) && ((BlockRockVariant) us.func_177230_c()).type == Rock.Type.DRY_GRASS)
                        spreader = Rock.Type.DRY_GRASS;

                    BlockRockVariant block = ((BlockRockVariant) current.func_177230_c());
                    world.func_175656_a(target, block.getVariant(block.type.getGrassVersion(spreader)).func_176223_P());
                }
            }
        }
    }

    public static boolean containsAnyOfCaseInsensitive(Collection<String> input, String... items)
    {
        Set<String> itemsSet = Arrays.stream(items).map(String::toLowerCase).collect(Collectors.toSet());
        return input.stream().map(String::toLowerCase).anyMatch(itemsSet::contains);
    }

    public static <T extends TileEntity> T getTE(IBlockAccess world, BlockPos pos, Class<T> aClass)
    {
        TileEntity te = world.func_175625_s(pos);
        if (!aClass.isInstance(te)) return null;
        //noinspection unchecked
        return (T) te;
    }

    public static String getEnumName(Enum<?> anEnum)
    {
        return JOINER_DOT.join(TFCConstants.MOD_ID, "enum", anEnum.getDeclaringClass().getSimpleName(), anEnum).toLowerCase();
    }

    public static String getTypeName(IForgeRegistryEntry<?> type)
    {
        return JOINER_DOT.join(TFCConstants.MOD_ID, "types", type.getRegistryType().getSimpleName(), type.getRegistryName().func_110623_a()).toLowerCase();
    }

    @Nonnull
    public static ItemStack consumeItem(ItemStack stack, int amount)
    {
        if (stack.func_190916_E() <= amount) return ItemStack.field_190927_a;
        stack.func_190918_g(amount);
        return stack;
    }

    @Nonnull
    public static ItemStack consumeItem(ItemStack stack, EntityPlayer player, int amount)
    {
        return player.func_184812_l_() ? stack : consumeItem(stack, amount);
    }

    /**
     * Simple method to spawn items in the world at a precise location, rather than using InventoryHelper
     */
    public static void spawnItemStack(World world, BlockPos pos, ItemStack stack)
    {
        if (stack.func_190926_b())
            return;
        EntityItem entityitem = new EntityItem(world, pos.func_177958_n(), pos.func_177956_o(), pos.func_177952_p(), stack);
        world.func_72838_d(entityitem);
    }

    /**
     * Method for hanging blocks to check if they can hang. 11/10 description.
     * NOTE: where applicable, remember to still check if the blockstate allows for the specified direction!
     *
     * @param pos    position of the block that makes the check
     * @param facing the direction the block is facing. This is the direction the block should be pointing and the side it hangs ON, not the side it sticks WITH.
     *               e.g: a sign facing north also hangs on the north side of the support block
     * @return true if the side is solid, false otherwise.
     */
    public static boolean canHangAt(World worldIn, BlockPos pos, EnumFacing facing)
    {
        return worldIn.isSideSolid(pos.func_177972_a(facing.func_176734_d()), facing);
    }

    /**
     * Primarily for use in placing checks. Determines a solid side for the block to attach to.
     *
     * @param pos             position of the block/space to be checked.
     * @param possibleSides   a list/array of all sides the block can attach to. Enumfacing constants are recommended.
     *                        This MUST NOT contain null!
     * @param prefferedFacing this facing is checked first. It can be invalid or null.
     * @return Found facing or null is none is found. This is the direction the block should be pointing and the side it stick TO, not the side it sticks WITH.
     */
    public static EnumFacing getASolidFacing(World worldIn, BlockPos pos, EnumFacing[] possibleSides, @Nullable EnumFacing prefferedFacing)
    {
        return getASolidFacing(worldIn, pos, Arrays.asList(possibleSides), prefferedFacing);
    }

    public static EnumFacing getASolidFacing(World worldIn, BlockPos pos, Collection<EnumFacing> possibleSides, @Nullable EnumFacing prefferedFacing)
    {
        if (possibleSides.contains(prefferedFacing) && canHangAt(worldIn, pos, prefferedFacing)) return prefferedFacing;
        for (EnumFacing side : possibleSides)
            if (canHangAt(worldIn, pos, side)) return side;
        return null;
    }
}
